/* $Header: /home/yav/xpx/RCS/convert.c,v 1.40 1996/04/24 23:08:27 yav Exp $
 * convert image data
 * written by yav (UHD98984@pcvan.or.jp)
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "headers.h"
#include "xpx.h"
#include "work.h"
#include "infodef.h"
#define PUBLIC_CONVERT_C
#include "extern.h"

char rcsid_convert[] = "$Id: convert.c,v 1.40 1996/04/24 23:08:27 yav Exp $";
static unsigned char rbittbl[8] = {0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};
static char tilemask[8][8];

int init_tilemask()
{
  return 0;
}

int set_tilemask(p)
     char *p;
{
  memcpy(tilemask, p, sizeof(tilemask));
  return 0;
}

void imgupdate(w, h)
     int w;
     int h;
{
  set_image_size(w, h);
  image_update_all();
}

/* return pnm type number */
int read_pnm_header(fp, data, colorhints, nhints)
     FILE *fp;
     int *data;
     COL *colorhints;
     int *nhints;
{
  int i, result, r, g, b;
  int datan, maxcol, datan0;
  char buf[256], *p;
  
  if (fgets(buf, sizeof(buf), fp) == NULL)
    return 0;
  if (buf[0] != 'P')
    return 0;			/* Not pnm file */
  switch (buf[1]) {
  case '1':
    result = 1;
    datan0 = 2;
    data[2] = 1;
    break;
  case '2':
    result = 2;
    datan0 = 3;
    break;
  case '3':
    result = 3;
    datan0 = 3;
    break;
  case '4':
    result = 4;
    datan0 = 2;
    data[2] = 1;
    break;
  case '5':
    result = 5;
    datan0 = 3;
    break;
  case '6':
    result = 6;
    datan0 = 3;
    break;
  default:
    return 0;			/* Not supported pnm file */
  }
  maxcol = -1;
  datan = 0;
  while (datan < datan0 && fgets(buf, sizeof(buf), fp) != NULL) {
    if (buf[0] == '#') {
      if (colorhints != NULL && nhints != NULL) {
	if (sscanf(buf, "# color:%d %d %d %d", &i, &r, &g, &b) == 4) {
	  if (i < *nhints) {
	    (colorhints+i)->rgb.red = r;
	    (colorhints+i)->rgb.green = g;
	    (colorhints+i)->rgb.blue = b;
	    if (maxcol < i)
	      maxcol = i;
	  }
	}
      }
    } else {
      for (p = buf; (i = strtol(p, &p, 10)) != 0 && datan < 3; datan++)
	data[datan] = i;
    }
  }
  if (colorhints != NULL && nhints != NULL)
    *nhints = maxcol+1;
  return result;
}

static unsigned char bitpos;
static unsigned char rddata;
static int pnmmaxval;

int read_pnm_pixels(fp, t, cp, n)
     FILE *fp;
     int t;
     COL *cp;
     int n;
{
  int i, x, rgbsw;
  int c;
  unsigned char *p, buf[256];
  
  switch(t) {
  case 1:			/* P1 pbm ascii */
  case 2:			/* P2 pgm ascii */
  case 3:			/* P3 ppm ascii */
    x = rgbsw = 0;
    while (x < n) {
      i = 0;
      p = buf;
      for (;;) {
	c = fgetc(fp);
	if (c == EOF)
	  return x;
	else if (c == '\n')
	  break;
	else if (c == ' '||c == '\t') {
	  if (i)
	    break;
	} else {
	  *p++ = c;
	  i = 1;
	}
      }
      *p = '\0';
      i = strtol((char *)buf, (char **)&p, 10);
      if (p == buf)
	continue;
      i = (i * 255) / pnmmaxval;
      switch(t) {
      case 3:
	switch(rgbsw) {
	case 0:
	  cp->rgb.red = i;
	  break;
	case 1:
	  cp->rgb.green = i;
	  break;
	case 2:
	  cp->rgb.blue = i;
	  break;
	}
	if (++rgbsw >= 3) {
	  cp->rgb.padding = 0;
	  cp++;
	  rgbsw = 0;
	  x++;
	}
	break;
      case 1:
      case 2:
	cp->rgb.red = cp->rgb.green = cp->rgb.blue = i;
	cp->rgb.padding = 0;
	cp++;
	x++;
	break;
      }
    }
    return x;
  case 4:			/* P4 pbm raw */
    while (n--) {
      if (bitpos == 0x01) {
	i = fgetc(fp);
	if (i == EOF)
	  return 0;
	rddata = i;
	bitpos = 0x80;
      } else {
	bitpos >>= 1;
      }
      cp->rgb.red = cp->rgb.green = cp->rgb.blue = rddata & bitpos ? 0 : 255;
      cp++;
    }
    return 1;
  case 5:			/* P5 pgm raw */
    while (n) {
      i = sizeof(buf) > n ? n : sizeof(buf);
      if (!fread(buf, i, 1, fp))
	return 0;
      p = buf;
      x = i;
      while (x--) {
	cp->rgb.red = cp->rgb.green = cp->rgb.blue = (*p++ * 255) / pnmmaxval;
	cp->rgb.padding = 0;
	cp++;
      }
      n -= i;
    }
    return 1;
  case 6:			/* P6 ppm raw */
    while (n) {
      i = sizeof(buf)/3 > n ? n : sizeof(buf)/3;
      if (!fread(buf, i*3, 1, fp))
	return 0;
      p = buf;
      x = i;
      while (x--) {
	cp->rgb.red = (*p++ * 255) / pnmmaxval;
	cp->rgb.green = (*p++ * 255) / pnmmaxval;
	cp->rgb.blue = (*p++ * 255) / pnmmaxval;
	cp->rgb.padding = 0;
	cp++;
      }
      n -= i;
    }
    return 1;
  default:
    /* unknown pnm type ??? */
    return 0;
  }
  /* NOT REACHED */
}

static COL lastcol;
static int lastcol_pixel;

int color_pixel(col)
     COL *col;
{
  int i;
  COL *tbl;
  
  if (lastcol_pixel >= 0) {
    if (memcmp(&lastcol.rgb, &col->rgb, 3) == 0)
      return lastcol_pixel;
  }
  tbl = color_buf;
  for (i = 0; i < MAXCOLSET*MAXPALSET; i++) {
    if (memcmp(&col->rgb, &tbl->rgb, 3) == 0 &&
	(coltbl[i] & (CM_ALLOCED|CM_REQUEST))) {
      lastcol = *col;
      return lastcol_pixel = i;
    }
    tbl++;
  }
  return -1;
}

typedef struct {
  int i;
  int n;
  int pn;
} CLN;

int colcnt(p, len, ctbl, ntbl, n, itbl)
     COL *p;
     int len;
     COL *ctbl;
     CLN *ntbl;
     int n;
     unsigned char *itbl;
{
  int i, x, max;
  
  max = 0;
  for (x = 0; x < len; x++) {
    i = 0;
    for (;;) {
      if (i >= max) {
	if (++max > n)
	  return -1;
	*(ctbl+i) = *p;
      }
      if (memcmp(&(ctbl+i)->rgb, &p->rgb, 3) == 0)
	break;
      i++;
    }
    *itbl++ = i;
    ++((ntbl+i)->n);
    p++;
  }
  return max;
}

int ncolcmp(p1, p2)
     CLN *p1;
     CLN *p2;
{
  return p2->n - p1->n;
}

int icolcmp(p1, p2)
     CLN *p1;
     CLN *p2;
{
  return p1->i - p2->i;
}

void slow_convert(px, w, h)
     COL *px;
     int w;
     int h;
{
  int i, n, len, x, y;
  unsigned char *itbl, *ip;
  COL ctbl[256];
  CLN ntbl[256];
  
  if (debug_mode)
    fprintf(stderr, "slow_convert %dx%d\n", w, h);
  len = w*h;
  itbl = (unsigned char *)malloc(len);
  for (i = 256; i; --i) {
    bzero(ctbl, sizeof(*ctbl)*i);
    bzero(ntbl, sizeof(*ntbl)*i);
    n = colcnt(px, len, ctbl, ntbl, i, itbl);
    if (n >= 0)
      break;
  }
  message("phase-1 %d colors.\n", n);
  for (i = 0; i < 256; i++)
    ntbl[i].i = i;
  qsort(ntbl, n, sizeof(*ntbl), ncolcmp);
  for (i = 0; i < n; i++)
    ntbl[i].pn = i;
  for (;;) {
    free_color_all();
    message("try %d colors.\n", n);
    for (i = 0; i < n; i++)
      request_color2(i, ctbl+ntbl[i].i);
    i = get_requested_colors();
    message("get %d/%d request colors.\n", i, n);
    if (i == n)
      break;
    n -= 16;
    if (n < 0) {
      message("!too few color!\n");
      free(itbl);
      return;
    }
  }
  qsort(ntbl, n, sizeof(*ntbl), icolcmp);
  ip = itbl;
  for (y = 0; y < h; y++) {
    for (x = 0; x < w; x++) {
      *(imgdata+y*imgmaxw+x) = ntbl[*ip++].pn;
    }
  }      
  free(itbl);
}

/* read portable bitmap/graymap/pixmap */
int rd_pnm(fp)
     FILE *fp;
{
  int c, i, n, f, x, y, pnmtype;
  int data[3], w, h;
  int colcnt;
  COL *px0, *px, *px1;
  
  colcnt = MAXPALSET*MAXCOLSET;
  pnmtype = read_pnm_header(fp, data, color_buf, &colcnt);
  if (pnmtype == 0)
    return 1;
  if (pnmtype == 1||pnmtype == 4)
    imgfiledepth = 1;
  if (debug_mode) {
    fprintf(stderr, "rd_pnm: width %d, height %d max %d\n",
	    data[0], data[1], data[2]);
    fprintf(stderr, "rd_pnm: %d color hints\n", colcnt);
  }

  free_color_all();
  /* magtoppm commented color information set */
  get_color(color_buf, colcnt);
  
  w = data[0];
  h = data[1];
  pnmmaxval = data[2];
  if (w > imgmaxw) {
    w = imgmaxw;
    message("!\nwidth %d trancate to %d.\n", data[0], w);
  }
  if (h > imgmaxh) {
    h = imgmaxh;
    message("!\nheight %d trancate to %d.\n", data[1], h);
  }
  px = (COL *)malloc(sizeof(*px)*w);
  if (!quick_convert)
    px1 = px0 = (COL *)malloc(sizeof(*px0)*w*h);
  lastcol_pixel = -1;
  for (y = 0; y < h; y++) {
    f = 1;
    bitpos = 0x01;
    for (n = data[0]; n; n -= i) {
      i = n > w ? w : n;
      read_pnm_pixels(fp, pnmtype, px, i);
      if (!f)
	continue;		/* skip x over pixels */
      if (quick_convert) {
	if (colcnt && pnmtype == 5) {
	  /* color infomation contained portable graymap */
	  for (x = 0; x < i; x++)
	    *(imgdata+y*imgmaxw+x) = (px+x)->rgb.red;
	} else {
	  for (x = 0; x < i; x++) {
	    c = color_pixel(px+x);
	    if (c < 0)
	      c = request_color(px+x);
	    *(imgdata+y*imgmaxw+x) = c;
	  }
	}
      } else {
	memcpy(px1, px, sizeof(*px)*i);
	px1 += i;
      }
      f = 0;
    }
  }
  if (!quick_convert) {
    slow_convert(px0, w, h);
    free(px0);
  } else {
    get_requested_colors();
  }
  free(px);
  imgupdate(w, h);
  return 0;
}

static char kiss_magic_number[] = "KiSS";

/* read KISS kcf */
int rd_kcf(fp)
     FILE *fp;
{
  int i, s, bpc, cc, col, set, preload;
  COL *cp;
  unsigned char *p, buf[4096];
  
  preload = strlen(kiss_magic_number);
  if (!fread(buf, preload, 1, fp))
    return 1;
  if (strncmp(buf, kiss_magic_number, preload) == 0) {
    /* read new style kcf header section */
    if (!fread(buf+preload, 32-preload, 1, fp))
      return 1;
    /* check color file mark */
    if (buf[4] != 0x10)
      return 1;
    bpc = buf[5];
    col = buf[8]|(buf[9]<<8);
    set = buf[10]|(buf[11]<<8);
    preload = 0;
    infomenu[INFOMENU_GSKCF].mode |= MenuModeMaskOn;
  } else {
    /* old style kcf */
    bpc = 12;
    col = 16;
    set = 10;
    infomenu[INFOMENU_GSKCF].mode &= ~MenuModeMaskOn;
  }
  redraw_window(infomenu[INFOMENU_GSKCF].win);
  cp = color_buf;
  cc = ((bpc+7)>>3)*col;
  for (s = 0; s < set; s++) {
    if (!fread(buf+preload, cc-preload, 1, fp))
      return 1;
    p = buf;
    for (i = 0; i < col; i++) {
      if (bpc == 12) {
	cp->rgb.red = (*p&0xf0)|(*p>>4);
	cp->rgb.blue = (*p&0x0f)|((*p&0x0f)<<4);
	p++;
	cp->rgb.green = (*p&0x0f)|((*p&0x0f)<<4);
	p++;
      } else {
	cp->rgb.red = *p++;
	cp->rgb.green = *p++;
	cp->rgb.blue = *p++;
      }
      cp++;
    }
    preload = 0;
  }
  free_color();
  get_color(color_buf+16*current_pal, col);
  image_update_col();
  return 0;
}

/* write KISS kcf */
int wr_kcf(fp)
     FILE *fp;
{
  int i, s, bpc, col, set, cc;
  COL *cp;
  unsigned char *p, buf[4096];
  
  /* old style kcf spec set */
  bpc = 12;			/* red:4 blue:4 green:4 = 12bits/color */
  col = 16;
  set = 10;
  if (infomenu[INFOMENU_GSKCF].mode & MenuModeMaskOn) {
    bpc = 24;			/* red:8 green:8 blue:8 = 24bits/color */
    bzero(buf, 32);
    memcpy(buf, kiss_magic_number, strlen(kiss_magic_number));
    buf[4] = 0x10;		/* color file mark */
    buf[5] = bpc;
    buf[8] = col;
    buf[9] = col>>8;
    buf[10] = set;
    buf[11] = set>>8;
    if (!fwrite(buf, 32, 1, fp))
      return 1;
  }
  cp = color_buf;
  cc = ((bpc+7)>>3)*col;
  for (s = 0; s < set; s++) {
    p = buf;
    for (i = 0; i < col; i++) {
      if (bpc == 12) {
	*p++ = (cp->rgb.red & 0xf0)|(cp->rgb.blue >> 4);
	*p++ = cp->rgb.green >> 4;
      } else {
	*p++ = cp->rgb.red;
	*p++ = cp->rgb.green;
	*p++ = cp->rgb.blue;
      }
      cp++;
    }
    if (!fwrite(buf, cc, 1, fp))
      return 1;
  }
  return 0;
}

/* read KISS cel */
int rd_cel(fp)
     FILE *fp;
{
  int c, x, y, w, h, preload, bpp;
  int dataw, datah;
  int ofsx, ofsy;
  unsigned char buf[4096];
  
  preload = strlen(kiss_magic_number);
  if (!fread(buf, preload, 1, fp))
    return 1;
  if (strncmp(buf, kiss_magic_number, preload) == 0) {
    /* read new style cel file header section */
    if (!fread(buf+preload, 32-preload, 1, fp))
      return 1;
    /* check cel file mark */
    if (buf[4] != 0x20)
      return 1;
    bpp = buf[5];
    w = buf[8]|(buf[9]<<8);
    h = buf[10]|(buf[11]<<8);
    ofsx = buf[12]|(buf[13]<<8);
    ofsy = buf[14]|(buf[15]<<8);
    infomenu[INFOMENU_GSCEL].mode |= MenuModeMaskOn;
  } else {
    bpp = 4;
    w = buf[0]|(buf[1]<<8);
    h = buf[2]|(buf[3]<<8);
    ofsx = ofsy = 0;
    infomenu[INFOMENU_GSCEL].mode &= ~MenuModeMaskOn;
  }
  redraw_window(infomenu[INFOMENU_GSCEL].win);
  imgfiledepth = bpp;
  dataw = w;
  datah = h;
  if (ofsx+w > imgmaxw) {
    w = imgmaxw-ofsx;
    message("!\nwidth %d trancate to %d.\n", dataw+ofsx, w+ofsx);
  }
  if (ofsy+h > imgmaxh) {
    h = imgmaxh-ofsy;
    message("!\nheight %d trancate to %d.\n", datah+ofsy, h+ofsy);
  }
  for (y = 0; y < ofsy; y++) {
    for (x = 0; x < ofsx+w; x++) {
      *(imgdata+y*imgmaxw+x) = 0;
    }
  }
  for (y = ofsy; y < ofsy+h; y++) {
    if (!fread(buf, bpp==4?((dataw+1)>>1):dataw, 1, fp))
      return 1;
    for (x = 0; x < ofsx; x++)
      *(imgdata+y*imgmaxw+x) = 0;
    for (x = 0; x < w; x++) {
      if (bpp == 4) {
	c = buf[x>>1];
	if ((x & 1) == 0)
	  c >>= 4;
	else
	  c &= 0x0f;
      } else {
	c = buf[x];
      }
      *(imgdata+y*imgmaxw+ofsx+x) = c;
    }
  }
  hotspotx = ofsx;
  hotspoty = ofsy;
  imgupdate(ofsx+w, ofsy+h);
  return 0;
}

int get_celofs_y()
{
  int x, y;
  
  for (y = 0; y < imgfileh; y++) {
    for (x = 0; x < imgfilew; x++) {
      if (*(imgdata+y*imgmaxw+x) != 0)
	return y;
    }
  }
  return 0;
}

int get_celofs_x()
{
  int x, y;
  
  for (x = 0; x < imgfilew; x++) {
    for (y = 0; y < imgfileh; y++) {
      if (*(imgdata+y*imgmaxw+x) != 0)
	return x;
    }
  }
  return 0;
}

/* write KISS cel */
int wr_cel(fp)
     FILE *fp;
{
  int c, x, y, w, h, bpp, headersize;
  unsigned char buf[4096];
  
  w = imgfilew;
  h = imgfileh;
  bpp = 4;			/* 4bits/pixel */
  if (infomenu[INFOMENU_GSCEL].mode & MenuModeMaskOn) {
    headersize = 32;
    bzero(buf, headersize);
    memcpy(buf, kiss_magic_number, strlen(kiss_magic_number));
    hotspotx = get_celofs_x();
    hotspoty = get_celofs_y();
    w -= hotspotx;
    h -= hotspoty;
    buf[4] = 0x20;		/* cel file mark */
    buf[5] = bpp;
    buf[8] = w;
    buf[9] = w >> 8;
    buf[10] = h;
    buf[11] = h >> 8;
    buf[12] = hotspotx;
    buf[13] = hotspotx >> 8;
    buf[14] = hotspoty;
    buf[15] = hotspoty >> 8;
  } else {
    headersize = 4;
    hotspotx = hotspoty = 0;
    buf[0] = w;
    buf[1] = w >> 8;
    buf[2] = h;
    buf[3] = h >> 8;
  }
  if (!fwrite(buf, headersize, 1, fp))
    return 1;
  for (y = hotspoty; y < hotspoty+h; y++) {
    for (x = 0; x < w; x++) {
      c = *(imgdata+y*imgmaxw+hotspotx+x);
      if (bpp == 4) {
	if ((x & 1) == 0)
	  buf[x>>1] = c<<4;
	else
	  buf[x>>1] |= c;
      } else {
	buf[x] = c;
      }
    }
    if (!fwrite(buf, bpp==4?((w+1)>>1):w, 1, fp))
      return 1;
  }
  return 0;
}

/* write portable pixmap */
int wr_ppm(fp)
     FILE *fp;
{
  int i, x, y;
  COL *cp;
  unsigned char c, buf[3];
  
  fprintf(fp, "P6\n");
  fprintf(fp, "#Saved by %s %s\n", myname, xpx_version);
  /* color information */
  cp = color_buf;
  for (i = 0; i < alloced_color; i++) {
    fprintf(fp, "# color:%3d %3d %3d %3d\n", i,
	    cp->rgb.red, cp->rgb.green, cp->rgb.blue);
    cp++;
  }
  fprintf(fp, "%d %d\n", imgfilew, imgfileh);
  fprintf(fp, "255\n");
  
  for (y = 0; y < imgfileh; y++) {
    for (x = 0; x < imgfilew; x++) {
      c = *(imgdata+y*imgmaxw+x);
      buf[0] = color_buf[c].rgb.red;
      buf[1] = color_buf[c].rgb.green;
      buf[2] = color_buf[c].rgb.blue;
      if (!fwrite(buf, 3, 1, fp))
	return 1;
    }
  }
  return 0;
}

int write_xbm_out(fp, data, datalen, end)
     FILE *fp;
     unsigned char *data;
     int datalen;
     int end;
{
  int i;
  char *p;
  char buf[256];
  
  strcpy(buf, "  ");
  p = buf+2;
  for (i = 0; i < datalen; i++) {
    sprintf(p, " 0x%02x,", *data++);
    p += strlen(p);
  }
  if (end)
    strcpy(p-1, "};");
  strcat(buf, "\n");
  return fputs(buf, fp);
}

/* write X11 bitmap */
int wr_xbm(fp, name)
     FILE *fp;
     char *name;
{
  int i, x, y, w, h;
  int outcount;
  unsigned char c;
  unsigned char outbuf[12];
  char namebuf[MAXPATH], *p;
  
  strcpy(namebuf, name);
  p = basefilename(namebuf);
  w = imgfilew;
  h = imgfileh;
  fprintf(fp, "#define %s_width %d\n", p, w);
  fprintf(fp, "#define %s_height %d\n", p, h);
  if (hotspotx >= 0 && hotspoty >= 0) {
    fprintf(fp, "#define %s_x_hot %d\n", p, hotspotx);
    fprintf(fp, "#define %s_y_hot %d\n", p, hotspoty);
  }
  fprintf(fp, "static unsigned char %s_bits[] = {\n", p);
  outcount = 0;
  for (y = 0; y < h; y++) {
    for (x = 0; x < w; x += 8) {
      c = 0;
      for (i = 0; i < 8; i++) {
	if ((x + i < w)&& *(imgdata+y*imgmaxw+x+i))
	  c |= rbittbl[i];
      }
      if (outcount >= sizeof(outbuf)) {
	write_xbm_out(fp, outbuf, outcount, 0);
	outcount = 0;
      }
      outbuf[outcount++] = c;
    }
  }
  write_xbm_out(fp, outbuf, outcount, 1);
  return 0;
}

int strcmp_tail(p, tail)
     char *p;
     char *tail;
{
  int pl, tl;
  
  tl = strlen(tail);
  pl = strlen(p);
  if (pl < tl)
    return 1;
  return strcmp(p+pl-tl, tail);
}

/* read X11 bitmap */
int rd_xbm(fp)
     FILE *fp;
{
  int i, x, y, w, h, c, hotx, hoty;
  char buf[1024], *p;
  char namebuf[MAXPATH];
  
  w = h = 0;
  hotx = hoty = -1;
  while (fgets(buf, sizeof(buf), fp) != NULL) {
    if (strncmp(buf, "static", 6) == 0||
	strncmp(buf, "unsigned", 8) == 0||
	strncmp(buf, "char", 4) == 0)
      break;
    if (sscanf(buf, "#define %s %d", namebuf, &c) == 2) {
      if (strcmp_tail(namebuf, "width") == 0)
	w = c;
      else if (strcmp_tail(namebuf, "height") == 0)
	h = c;
      else if (strcmp_tail(namebuf, "x_hot") == 0)
	hotx = c;
      else if (strcmp_tail(namebuf, "y_hot") == 0)
	hoty = c;
    }
  }
  if (!w || !h)
    return 1;
  y = x = 0;
  p = buf;
  *p = '\0';
  for (;;) {
    while (*p && (*p == ' ' || *p == '\t' || *p == ','))
      p++;
    if (*p == '\0' || *p == '\n') {
      if (fgets(buf, sizeof(buf), fp) == NULL)
	break;
      p = buf;
    }
    if (sscanf(p, "0x%02x", &c)) {
      p += 4;
      for (i = 0; i < 8; i++) {
	if (x+i < w)
	  *(imgdata+y*imgmaxw+x+i) = ((c & rbittbl[i]) ? 1 : 0);
      }
      x += 8;
      if (x >= w) {
	if (++y >= h)
	  break;
	x = 0;
      }
    } else {
      p++;
    }
  }
  imgfiledepth = 1;
  hotspotx = hotx;
  hotspoty = hoty;
  imgupdate(w, h);
  return 0;
}

/* set XImage from imgdata */
void setpixel_sub2(x0, y0, pixel)
     int x0;
     int y0;
     unsigned long pixel;
{
  int x, y, x1, y1;
  
  x0 *= zoomfactor;
  y0 *= zoomfactor;
  x1 = x0+zoomfactor;
  y1 = y0+zoomfactor;
  for (y = y0; y < y1; y++)
    for (x = x0; x < x1; x++)
      XPutPixel(img, x, y, pixel);
}

/* set XImage from imgdata */
void setpixel_sub(x0, y0, pixel)
     int x0;
     int y0;
     unsigned long pixel;
{
  int x, y, x1, y1;
  
  if (vimg != NULL)
    XPutPixel(vimg, x0, y0,  pixel);
  x0 *= zoomfactor;
  y0 *= zoomfactor;
  x1 = x0+zoomfactor;
  y1 = y0+zoomfactor;
  for (y = y0; y < y1; y++)
    for (x = x0; x < x1; x++)
      XPutPixel(img, x, y, pixel);
}

/* set view image */
#define setpixel_subv(x0,y0,px) XPutPixel(vimg,(x0),(y0),(px))

void setpixelcolor(x, y, col)
     int x;
     int y;
     int col;
{
  unsigned char *p;
  
  /* check masked color */
  p = imgdata+y*imgmaxw+x;
  if (!*(colmask + current_pal*MAXCOLSET + *p) && tilemask[y&7][x&7]) {
    *p = col;
    setpixel_sub2(x, y, pixtbl[col]);
    if (vimg != NULL)
      setpixel_subv(x, y, pixtbl[col]);
  }
}

void gen_image_subv(x0, y0, w, h)
     int x0;
     int y0;
     int w;
     int h;
{
  int x, y;
  unsigned char *ip;
  
  for (y = y0; y < y0+h; y++) {
    ip = imgdata+y*imgmaxw+x0;
    for (x = x0; x < x0+w; x++) {
      setpixel_subv(x, y, pixtbl[*ip++]);
    }
  }
}

/* set XImage from imgdata */
void gen_image_sub(x0, y0, w, h)
     int x0;
     int y0;
     int w;
     int h;
{
  int x, y, xs, ys, xe, ye, zx, zy;
  unsigned char *ip;
  unsigned char pix;
  
  xs = x0 * zoomfactor;
  ys = y0 * zoomfactor;
  xe = xs + w * zoomfactor;
  ye = ys + h * zoomfactor;
  for (y = ys; y < ye; y += zoomfactor) {
    ip = imgdata+(y/zoomfactor)*imgmaxw+x0;
    for (x = xs; x < xe; x += zoomfactor) {
      pix = pixtbl[*ip++];
      for (zy = 0; zy < zoomfactor; zy++)
	for (zx = 0; zx < zoomfactor; zx++)
	  XPutPixel(img, x+zx, y+zy, pix);
    }
  }
}

void update_area(x, y, w, h)
     int x;
     int y;
     int w;
     int h;
{
  gen_image_sub(x, y, w, h);
  if (vimg != NULL)
    gen_image_subv(x, y, w, h);
}

void fill_area2(dstx, dsty, w, h, c)
     int dstx;
     int dsty;
     int w;
     int h;
     int c;
{
  int x, y, xe, ye;
  unsigned char *p;
  char *mp;
  
  mp = colmask + current_pal*MAXCOLSET;
  xe = dstx + w;
  ye = dsty + h;
  for (y = dsty; y < ye; y++) {
    p = imgdata+y*imgmaxw+dstx;
    for (x = dstx; x < xe; x++) {
      if (!*(mp+*p))
	*p = c;
      p++;
    }
  }
  update_area(dstx, dsty, w, h);
}

void fill_area(dstx, dsty, w, h, c)
     int dstx;
     int dsty;
     int w;
     int h;
     int c;
{
  int x, y, xe, ye;
  unsigned char *p;
  char *mp;
  
  mp = colmask + current_pal*MAXCOLSET;
  xe = dstx + w;
  ye = dsty + h;
  for (y = dsty; y < ye; y++) {
    p = imgdata+y*imgmaxw+dstx;
    for (x = dstx; x < xe; x++) {
      if (!*(mp+*p) && tilemask[y&7][x&7])
	*p = c;
      p++;
    }
  }
  update_area(dstx, dsty, w, h);
}

void copy_area(dstx, dsty, srcx, srcy, w, h)
     int dstx;
     int dsty;
     int srcx;
     int srcy;
     int w;
     int h;
{
  int x, y;
  int lastnocopypixel;
  unsigned char *srcp, *dstp;
  char *mp;

  lastnocopypixel = nocopypixel;
  if (!nocopypixel_mode)
    nocopypixel = -1;
  mp = colmask+current_pal*MAXPALSET;
  if (dsty*imgmaxw+dstx > srcy*imgmaxw+srcx) {
    for (y = h-1; y >= 0; y--) {
      dstp = imgdata+(dsty+y)*imgmaxw+dstx;
      srcp = imgdata+(srcy+y)*imgmaxw+srcx;
      for (x = w-1; x >= 0; x--) {
	if (!*(mp+*(dstp+x)) && *(srcp+x) != nocopypixel)
	  *(dstp+x) = *(srcp+x);
      }
    }
  } else {
    for (y = 0; y < h; y++) {
      dstp = imgdata+(dsty+y)*imgmaxw+dstx;
      srcp = imgdata+(srcy+y)*imgmaxw+srcx;
      for (x = 0; x < w; x++) {
	if (!*(mp+*(dstp+x)) && *(srcp+x) != nocopypixel)
	  *(dstp+x) = *(srcp+x);
      }
    }
  }
  update_area(dstx, dsty, w, h);
  nocopypixel = lastnocopypixel;
}


void gen_image_col(x0, y0, w, h)
     int x0;
     int y0;
     int w;
     int h;
{
  int x, y;
  unsigned char *ip;
  
  for (y = y0; y < y0+h; y++) {
    ip = imgdata+y*imgmaxw+x0;
    for (x = x0; x < x0+w; x++) {
      if (coltbl[*ip] & CM_UPDATING) {
	coltbl[*ip] |= CM_UPDATED;
	setpixel_sub2(x, y, pixtbl[*ip++]);
      }
    }
  }
  if (vimg != NULL) {
    for (y = y0; y < y0+h; y++) {
      ip = imgdata+y*imgmaxw+x0;
      for (x = x0; x < x0+w; x++) {
	if (coltbl[*ip] & CM_UPDATING) {
	  coltbl[*ip] |= CM_UPDATED;
	  setpixel_subv(x, y, pixtbl[*ip++]);
	}
      }
    }
  }
  for (x = 0; x < MAXPALSET*MAXCOLSET; x++) {
    if (coltbl[x] & CM_UPDATED)
      coltbl[x] &= ~(CM_UPDATED|CM_UPDATING);
  }
}

/* End of file */
