/* $Header: /home/yav/xpx/RCS/image.c,v 1.18 1996/03/26 15:39:13 yav Exp $
 * xpx image control
 * written by yav (UHD98984@pcvan.or.jp)
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#include "headers.h"
#include "xpx.h"
#include "work.h"
#include "scrldef.h"
#include "infodef.h"
#include "cursdef.h"
#include "editdef.h"
#define PUBLIC_IMAGE_C
#include "extern.h"

#ifndef IMGWINWIDTHMAX
#define IMGWINWIDTHMAX 1024
#endif
#ifndef IMGWINHEIGHTMAX
#define IMGWINHEIGHTMAX 1024
#endif

char rcsid_image[] = "$Id: image.c,v 1.18 1996/03/26 15:39:13 yav Exp $";

AREA area0;
Bool area_flag = False;
Bool mark_flag = False;

static AREA lastarea;
static AREA current_area;
static int mark_cnt = 0;
static int *mark_pts = NULL;

void get_img_window_size()
{
  XWindowAttributes attr;

  XGetWindowAttributes(dsp, win, &attr);
  imgwinwidth = attr.width - imgwx - BDW*2;
  imgwinheight = attr.height - imgwy - BDW*2;
  if (scrollbarmode) {
    imgwinwidth -= IMGSCRLW + BDW*3 + 3;
    imgwinheight -= IMGSCRLW + BDW*3 + 3;
  }
  if (imgwinwidth > IMGWINWIDTHMAX)
    imgwinwidth = IMGWINWIDTHMAX;
  if (imgwinheight > IMGWINHEIGHTMAX)
    imgwinheight = IMGWINHEIGHTMAX;
}

int create_img_window()
{
  int i;
  unsigned long mask;
  XSetWindowAttributes attr;
  
  mark_pts = (int *)malloc(sizeof(int)*8*(imgmaxw+imgmaxh));
  if (mark_pts == NULL)
    return error("create_img_window: more core!");
  get_img_window_size();
  imgwin = XCreateSimpleWindow(dsp, win, imgwx, imgwy,
			       imgwinwidth, imgwinheight, BDW, fg, bg);
  attr.bit_gravity = NorthWestGravity;
  mask = CWBitGravity;
  if (backingstoremode) {
    switch(backingstoremode) {
    case 1:
      i = WhenMapped;
      break;
    case 2:
      i = Always;
      break;
    default:
      i = backingstoremode;
      break;
    }
    attr.backing_store = i;
    mask |= CWBackingStore;
  }
  XChangeWindowAttributes(dsp, imgwin, mask, &attr);
  XSelectInput(dsp, imgwin, ExposureMask|
	       ButtonPressMask|ButtonReleaseMask|ButtonMotionMask);
  XMapRaised(dsp, imgwin);
#if 0
  csource = XCreateBitmapFromData(dsp, win,
				  curs_bits, curs_width, curs_height);
  cmask = XCreateBitmapFromData(dsp, win, cursmask_bits,
				cursmask_width, cursmask_height);
  XParseColor(dsp, DefaultColormap(dsp, scr), "Black", &cfore);
  XParseColor(dsp, DefaultColormap(dsp, scr), "White", &cback);
  curs = XCreatePixmapCursor(dsp, csource, cmask, &cfore, &cback,
			     curs_x_hot, curs_y_hot);
  /*  curs = XCreateFontCursor(dsp, XC_tcross); */
  XDefineCursor(dsp, imgwin, curs);
#endif
  return 0;
}

int image_alloc()
{
  imgdata = (unsigned char *)malloc(imgmaxw*imgmaxh);
  if (imgdata == NULL)
    return 1;
  img = XCreateImage(dsp, DefaultVisual(dsp, scr), DefaultDepth(dsp, scr),
		     ZPixmap, 0, None,
		     imgmaxw*maxzoom, imgmaxh*maxzoom,
		     32, 0);
  if (img == NULL)
    return 1;
  img->data = (char *)malloc(img->bytes_per_line * img->height);
  if (img->data == NULL)
    return 1;
  return 0;
}

void init_image()
{
  if (!alloced_color)
    get_color(color_buf+MAXCOLSET*current_pal, MAXCOLSET);
  bzero(img->data, img->bytes_per_line * img->height);
}

void resize_img_window()
{
  get_img_window_size();
  XResizeWindow(dsp, imgwin, imgwinwidth, imgwinheight);
  resize_scrl();
  resize_scroll_slider();
}

int set_image_size(w, h)
     int w;
     int h;
{
  int r;

  r = 0;
  if (w >= 0) {
    imgfilew = w;
    r = 1;
  }
  if (h >= 0) {
    imgfileh = h;
    r |= 2;
  }
  info_imgsize_set();
  resize_scroll_slider();
  redraw_window(imgwin);
  return r;
}

void draw_grid(x0, y0, w0, h0)
     int x0;
     int y0;
     int w0;
     int h0;
{
  int n, x, y, w, h;
  int ox, oy;

  n = grid_size*zoomfactor;
  ox = imgofsx + (grid_size-grid_ofsx) * zoomfactor;
  oy = imgofsy + (grid_size-grid_ofsy) * zoomfactor;
  w = w0;
  x = x0;
  if (x & 1) {
    x++;
    w--;
  }
  if (w > 0 && h0 > 0) {
    for (y = y0; y < y0 + h0; y++) {
      if (((oy + y)%n) == 0)
	XDrawLine(dsp, imgwin, gridgc, x, y, x+w-1, y);
    }
  }
  h = h0;
  y = y0;
  if (y & 1) {
    y++;
    h--;
  }
  if (w0 > 0 && h > 0) {
    for (x = x0; x < x0 + w0; x++) {
      if (((ox + x)%n) == 0)
	XDrawLine(dsp, imgwin, gridgc, x, y, x, y+h-1);
    }
  }
}

Bool inarea(x, w, n1, n2)
     int x;
     int w;
     int n1;
     int n2;
{
  if (x < n1)
    return (x+w >= n1);
  return x < n2;
}

void calc_img_area(x, y, w, h)
     int *x;
     int *y;
     int *w;
     int *h;
{
  *w *= zoomfactor;
  *h *= zoomfactor;
  *x = *x * zoomfactor - imgofsx;
  *y = *y * zoomfactor - imgofsy;
}

void little_big(p1, p2)
     int *p1;
     int *p2;
{
  int i;
  
  if (*p1 > *p2) {
    i = *p1;
    *p1 = *p2;
    *p2 = i;
  }
}

void calc_area(x0, y0, x1, y1)
     int *x0;
     int *y0;
     int *x1;
     int *y1;
{
  little_big(x0, x1);
  little_big(y0, y1);
  *x1 -= *x0;
  *y1 -= *y0;
  calc_img_area(x0, y0, x1, y1);
}

void area_expose(x0, y0, w0, h0)
     int x0;
     int y0;
     int w0;
     int h0;
{
  AREA ar;

  ar = current_area;
  calc_area(&ar.x, &ar.y, &ar.w, &ar.h);
  if ((inarea(ar.x, ar.w, x0, x0+w0) ||
       inarea(ar.y, ar.h, y0, y0+h0)))
    XDrawRectangle(dsp, imgwin, areagc, ar.x, ar.y, ar.w, ar.h);
}

void mark_pts_on(x0, y0, w0, h0)
     int x0;
     int y0;
     int w0;
     int h0;
{
  int i, x, y;
  int *p;

  p = mark_pts;
  i = mark_cnt;
  while (i--) {
    x = *p++ * zoomfactor - imgofsx;
    y = *p++ * zoomfactor - imgofsy;
    if (x0 <= x && x < x0 + w0 && y0 <= y && y < y0 + h0)
      XDrawPoint(dsp, imgwin, areagc, x, y);
  }
}

void img_expose(x, y, w, h)
     int x;
     int y;
     int w;
     int h;
{
  int i;
  int hx, hy;
  
  if ((imgofsx+x >= imgfilew*zoomfactor)||
      (imgofsy+y >= imgfileh*zoomfactor)||
      (imgofsx+x+w < 0)||
      (imgofsy+y+h < 0)) {
    if (x >= 0 && x+w <= imgwinwidth && y >= 0 && y+h <= imgwinheight)
      XClearArea(dsp, imgwin, x, y, w, h, False);
    return;
  }
  i = -imgofsx - x;
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y, i, h, False);
    x += i;
    w -= i;
  }
  i = -imgofsy - y;
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y, w, i, False);
    y += i;
    h -= i;
  }
  i = x + w - (imgfilew*zoomfactor - imgofsx);
  if (i > 0) {
    XClearArea(dsp, imgwin, x+w-i, y, i, h, False);
    w -= i;
  }
  i = y + h - (imgfileh*zoomfactor - imgofsy);
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y+h-i, w, i, False);
    h -= i;
  }
  if (w && h) {
    XPutImage(dsp, imgwin, gc0, img, x+imgofsx, y+imgofsy, x, y, w, h);
    if (grid_mode)
      draw_grid(x, y, w, h);
    if (area_flag)
      area_expose(x, y, w, h);
    if (mark_flag)
      mark_pts_on(x, y, w, h);
    hx = hotspotx*zoomfactor-imgofsx;
    hy = hotspoty*zoomfactor-imgofsy;
    if (infomenu[INFOMENU_HOTSPOTCENTER].mode & MenuModeMaskOn) {
      hx += zoomfactor/2;
      hy += zoomfactor/2;
    }
    if (y <= hy && hy < y+h)
      XDrawLine(dsp, imgwin, grid0gc, x, hy, x+w-1, hy);
    if (x <= hx && hx < x+w)
      XDrawLine(dsp, imgwin, grid0gc, hx, y, hx, y+h-1);
  }
}

void image_update_all_sub()
{
  img_expose(0, 0, imgwinwidth, imgwinheight);
  if (view_sw & 1)
    redraw_window(viewwin);
}

void pre_image_udpate()
{
  cursor_busy();
  make_view_image();
}

void image_update_all()
{
  pre_image_udpate();
  gen_image_sub(0, 0, imgfilew, imgfileh);
  image_update_all_sub();
}

void image_update_col()
{
  pre_image_udpate();
  gen_image_col(0, 0, imgfilew, imgfileh);
  image_update_all_sub();
}

void color_update_all()
{
  free_color();
  get_color(color_buf+MAXCOLSET*current_pal, MAXCOLSET);
  if (alloced_color > alloced_private_color)
    image_update_col();
}

void img_scroll_redraw(dx, dy)
     int dx;
     int dy;
{
  if (imgwin_noexpose) {
    XCopyArea(dsp, imgwin, imgwin, gc0,
	      (dx<0?0:dx), (dy<0?0:dy),
	      imgwinwidth-(dx<0?-dx:dx), imgwinheight-(dy<0?-dy:dy),
	      (dx<0?-dx:0), (dy<0?-dy:0));
    imgwin_noexpose = False;
    if (dx)
      img_expose((dx<0?0:imgwinwidth-dx), 0, (dx<0?-dx:dx), imgwinheight);
    if (dy)
      img_expose(0, (dy<0?0:imgwinheight-dy), imgwinwidth, (dy<0?-dy:dy));
  } else {
    img_expose(0, 0, imgwinwidth, imgwinheight);
  }
}

static int center_x = 0;
static int center_y = 0;

void img_scroll_sub(dx, dy)
     int *dx;
     int *dy;
{
  if (infomenu[INFOMENU_SCROLLSTOP].mode & MenuModeMaskOn) {
    if (*dx) {
      if (imgofsx + *dx + imgwinwidth >= imgfilew*zoomfactor)
	*dx = imgfilew*zoomfactor - (imgofsx + imgwinwidth);
      if (imgofsx + *dx < 0)
	*dx = -imgofsx;
    }
    if (*dy) {
      if (imgofsy + *dy + imgwinheight >= imgfileh*zoomfactor)
	*dy = imgfileh*zoomfactor - (imgofsy + imgwinheight);
      if (imgofsy + *dy < 0)
	*dy = -imgofsy;
    }
  }
  if (!*dx && !*dy)
    return;
  if (*dx) {
    imgofsx += *dx;
    move_hslider();
    redraw_window(cursmenu[CURS_TOPX].win);
  }
  if (*dy) {
    imgofsy += *dy;
    move_vslider();
    redraw_window(cursmenu[CURS_TOPY].win);
  }
  center_x = (imgofsx + imgwinwidth/2) / zoomfactor;
  center_y = (imgofsy + imgwinheight/2) / zoomfactor;
}

void img_scroll(dx, dy)
     int dx;
     int dy;
{
  img_scroll_sub(&dx, &dy);
  img_scroll_redraw(dx, dy);
}

void img_scroll_posset(newx, newy)
     int newx;
     int newy;
{
  int dx, dy;
  
  dx = newx-imgofsx;
  dy = newy-imgofsy;
  img_scroll_sub(&dx, &dy);
}

void img_scroll_posreset()
{
  int i, x, y;
  
  x = center_x;
  y = center_y;
  i = imgfilew - imgwinwidth/2/zoomfactor;
  if (x > i)
    x = i;
  i = imgfileh - imgwinheight/2/zoomfactor;
  if (y > i)
    y = i;
  if (x < 0)
    x = 0;
  if (y < 0)
    y = 0;
  x = (x * zoomfactor) - imgwinwidth/2;
  y = (y * zoomfactor) - imgwinheight/2;
  img_scroll_posset(x, y);
  resize_scroll_slider();
}

void imgscrl_background()
{
  int dx, dy, ax, ay;

  ax = scroll_ax;
  ay = scroll_ay;
  /* scroll bar press? */
  if (scrlbar[SCRLBAR_TOP].mode & MenuModeMaskOn ||
      vslider[VSLIDER_UP].mode & MenuModeMaskOn)
    ay = -1;
  if (scrlbar[SCRLBAR_BOTTOM].mode & MenuModeMaskOn ||
      vslider[VSLIDER_DOWN].mode & MenuModeMaskOn)
    ay = 1;
  if (scrlbar[SCRLBAR_LEFT].mode & MenuModeMaskOn ||
      hslider[HSLIDER_LEFT].mode & MenuModeMaskOn)
    ax = -1;
  if (scrlbar[SCRLBAR_RIGHT].mode & MenuModeMaskOn ||
      hslider[HSLIDER_RIGHT].mode & MenuModeMaskOn)
    ax = 1;
  /* set default background scroll */
  scroll_dx += ax;
  if (scroll_dx > scroll_speed)
    scroll_dx = scroll_speed;
  else if (scroll_dx < -scroll_speed)
    scroll_dx = -scroll_speed;
  scroll_dy += ay;
  if (scroll_dy > scroll_speed)
    scroll_dy = scroll_speed;
  else if (scroll_dy < -scroll_speed)
    scroll_dy = -scroll_speed;
  dx = scroll_dx;
  dy = scroll_dy;
  if (dx || dy)
    img_scroll(dx, dy);
}


void norm_area(x0, y0, x1, y1)
     int *x0;
     int *y0;
     int *x1;
     int *y1;
{
  little_big(x0, x1);
  little_big(y0, y1);
  *x1 -= *x0;
  *y1 -= *y0;
}

void area_exsub(x, y, w, h)
     int x;
     int y;
     int w;
     int h;
{
  img_expose(x, y, w+1, 1);
  img_expose(x, y, 1, h+1);
  img_expose(x+w, y, 1, h+1);
  img_expose(x, y+h, w+1, 1);
}

void erase_last_area()
{
  Bool af;

  af = area_flag;
  area_flag = False;
  calc_area(&lastarea.x, &lastarea.y, &lastarea.w, &lastarea.h);
  area_exsub(lastarea.x, lastarea.y, lastarea.w, lastarea.h);
  area_flag = af;
}

void copy_area_expose()
{
  AREA ar;
  
  if (area_flag) {
    erase_last_area();
    ar = lastarea = current_area = area0;
    calc_area(&ar.x, &ar.y, &ar.w, &ar.h);
    area_exsub(ar.x, ar.y, ar.w, ar.h);
  }
}

void max_area(rx0, ry0, rx1, ry1, n, p)
     int *rx0;
     int *ry0;
     int *rx1;
     int *ry1;
     int n;
     int *p;
{
  int x0, y0, x1, y1;
  
  x0 = x1 = *p++;
  y0 = y1 = *p++;
  while (--n) {
    if (*p < x0)
      x0 = *p;
    if (*p > x1)
      x1 = *p;
    p++;
    if (*p < y0)
      y0 = *p;
    if (*p > y1)
      y1 = *p;
    p++;
  }
  *rx0 = x0;
  *ry0 = y0;
  *rx1 = x1;
  *ry1 = y1;
}

void mark_pts_set(n, p)
     int n;
     int *p;
{
  memcpy(mark_pts, p, n*sizeof(int)*2);
  mark_cnt = n;
}

void mark_pts_expose()
{
  int xx0, xx1, yy0, yy1;

  if (mark_cnt > 0) {
    max_area(&xx0, &yy0, &xx1, &yy1, mark_cnt, mark_pts);
    img_expose(xx0*zoomfactor-imgofsx, yy0*zoomfactor-imgofsy,
	       (xx1-xx0+1)*zoomfactor, (yy1-yy0+1)*zoomfactor);
  }
}

/* End of file */
