/* $Header: /home/yav/xpx/RCS/mag.c,v 1.10 1996/04/23 10:58:05 yav Exp $
 * xpx save load MAG format
 * written by yav (UHD98984@pcvan.or.jp)
 */

#include <X11/Xlib.h>

#include "xpx.h"
#include "headers.h"
#include "work.h"
#define PUBLIC_MAG_C
#include "extern.h"


char rcsid_mag[] = "$Id: mag.c,v 1.10 1996/04/23 10:58:05 yav Exp $";
static char mag_magic_number[] = "MAKI02  ";
static char mag_default_user[] = ">\x93\xe4<";
static struct {int x,y;} magpostbl[16]={
  {0,0},{1,0},{2,0},{4,0},{0,1},{1,1},{0,2},{1,2},
  {2,2},{0,4},{1,4},{2,4},{0,8},{1,8},{2,8},{0,16}};
static long flagAsize;
static long flagBsize;
static long pixelsize;
static unsigned char *pixelbuf = NULL;
static unsigned char *flaga = NULL;
static unsigned char *flagb = NULL;


/* store short integer to char buffer as little-endian
 */
void setleshort(p, n)
     char *p;
     int n;
{
  *p++ = n;
  *p = n >> 8;
}

/* store long integer to char buffer as little-endian
 */
void setlelong(p, n)
     char *p;
     long n;
{
  *p++ = n;
  *p++ = n >> 8;
  *p++ = n >> 16;
  *p   = n >> 24;
}

/* write MAG palette data
 * Caution! Now supported only 16 colors format
 */
int write_mag_palette_section(fp)
     FILE *fp;
{
  int i;
  COL *cp;
  char buf[3*16], *p;
  
  cp = color_buf;
  p = buf;
  for (i = 0; i < 16; i++) {
    *p++ = cp->rgb.green;
    *p++ = cp->rgb.red;
    *p++ = cp->rgb.blue;
    cp++;
  }
  if (!fwrite(buf, 3*16, 1, fp))
    return 1;
  return 0;
}

int alloc_compress_work()
{
  int w;
  
  w = (imgfilew+7)/8;
  pixelbuf = (unsigned char *)malloc((imgfilew+1)/2*imgfileh);
  flaga = (unsigned char *)malloc(w*imgfileh + 1);
  flagb = (unsigned char *)malloc(w*imgfileh + 1);
  return pixelbuf == NULL || flaga == NULL || flagb == NULL;
}

void free_compress_work()
{
  if (pixelbuf != NULL) {
    free(pixelbuf);
    pixelbuf = NULL;
  }
  if (flaga != NULL) {
    free(flaga);
    flaga = NULL;
  }
  if (flagb != NULL) {
    free(flagb);
    flagb = NULL;
  }
}

int is_same_pixel(x, y, dd)
     int x;
     int y;
     int dd;
{
  int dx, dy;
  
  dx = x - magpostbl[dd].x * 4;
  dy = y - magpostbl[dd].y;
  if (dx < 0 || dy < 0)
    return 0;
  return memcmp(imgdata+imgmaxw*y+x, imgdata+imgmaxw*dy+dx, 4) == 0;
}

int search_same_pixel(x, y)
     int x;
     int y;
{
  int i, n;
  static int stbl[15] = {1,4,5,6,7,9,10,2,8,11,12,13,14,3,15};

  for (i = 0; i < 15; i++) {
    n = stbl[i];
    if (is_same_pixel(x, y, n))
      return n;
  }
  return 0;			/* not found */
}

/* compress MAG
 * imgdata imgmaxw x imgmaxh
 * effective area imgfilew x imgfileh
 */
int mag_compress()
{
  int x, y, n0, n1, w, bitpos;
  unsigned char *flagbuf, *pf;
  unsigned char *flagbuf2, *pf2; /* upper line flag cache */
  unsigned char *pi, *pp;
  unsigned char *pa, *pb;
  
  w = (imgfilew+7)/8;
  pf = flagbuf = (unsigned char *)malloc(w*imgfileh);
  if (flagbuf == NULL)
    return 1;
  flagbuf2 = (unsigned char *)malloc(w);
  if (flagbuf2 == NULL) {
    free(flagbuf);
    return 1;
  }
  if (alloc_compress_work()) {
    free(flagbuf);
    free_compress_work();
    return 1;
  }
  pp = pixelbuf;
  bzero(flagbuf2, w);
  for (y = 0; y < imgfileh; y++) {
    pi = imgdata+imgmaxw*y;
    pf2 = flagbuf2;
    for (x = 0; x < imgfilew; x += 8) {
      n0 = *pf2 >> 4;
      if (!n0 || !is_same_pixel(x, y, n0))
	n0 = search_same_pixel(x, y);
      if (!n0) {
	*pp++ = (*pi << 4) | *(pi+1);
	*pp++ = (*(pi+2) << 4) | *(pi+3);
      }
      pi += 4;
      n1 = *pf2 & 0x0f;
      if (!n1 || !is_same_pixel(x+4, y, n1))
	n1 = search_same_pixel(x+4, y);
      if (!n1) {
	*pp++ = (*pi << 4) | *(pi+1);
	*pp++ = (*(pi+2) << 4) | *(pi+3);
      }
      pi += 4;
      *pf2++ = *pf++ = (n0 << 4) | n1;
    }
  }      
  for (y = imgfileh-1; y > 0; y--) {
    pf = flagbuf+w*y;
    for (x = 0; x < w; x++)
      *pf++ ^= *(pf-w);
  }
  bitpos = 0x80;
  pf = flagbuf;
  pa = flaga;
  pb = flagb;
  for (y = 0; y < imgfileh; y++) {
    for (x = 0; x < w; x++) {
      if (*pf) {
	*pb++ = *pf;
	*pa |= bitpos;
      }
      if (!(bitpos >>= 1)) {
	pa++;
	*pa = 0;
	bitpos = 0x80;
      }
      pf++;
    }
  }
  if (bitpos != 0x80)
    pa++;
  flagAsize = pa - flaga;
  if (flagAsize & 1)
    flagAsize++;
  flagBsize = pb - flagb;
  if (flagBsize & 1)
    flagBsize++;
  pixelsize = pp - pixelbuf;
  free(flagbuf);
  free(flagbuf2);
  return 0;			/* success */
}

int wr_mag(fp)
     FILE *fp;
{
  int i, x, y, comp;
  unsigned char *p;
  long l, flagAofs, flagBofs, pixelofs;
  char *username, *comment;
  char buf[512];
  
  if (!imgfiledepth)
    imgfiledepth = 4;		/* If no depth info, treat 4 bit depth */
  if (imgfiledepth > 4) {
    message("!\n Supported only 16 color mag!\n");
    return 1;
  }
  comp = 1;
  username = (char *)getenv("USER");
  if (username == NULL)
    username = mag_default_user;
  sprintf(buf, "%sxpx  %-18.18s ", mag_magic_number, username);
  if (!fwrite(buf, strlen(buf), 1, fp))
    return 1;
  sprintf(buf, "Saved by %s %s", myname, xpx_version);
  comment = buf;
  if (!fwrite(comment, strlen(comment), 1, fp))
    return 1;
  if (!fwrite("\032", 1, 1, fp))
    return 1;
  if (comp) {
    if (mag_compress())
      comp = 0;		/* compression failed, select no compression */
  }
  if (!comp) {
    flagAsize = ((imgfilew+3)/4+1)/2;
    flagAsize *= imgfileh;
    if (flagAsize & 1)
      ++flagAsize;
    flagBsize = 0;
    pixelsize = ((imgfilew*imgfileh+3)/4) * 2;
  }
  flagAofs = 32 + 3*16;
  flagBofs = flagAofs + flagAsize;
  pixelofs = flagBofs + flagBsize;
  /* write header */
  bzero(buf, 32);
  setleshort(buf+4, imgposx);
  setleshort(buf+6, imgposy);
  setleshort(buf+8, imgposx+imgfilew-1);
  setleshort(buf+10, imgposy+imgfileh-1);
  setlelong(buf+12, flagAofs);
  setlelong(buf+16, flagBofs);
  setlelong(buf+20, flagBsize);
  setlelong(buf+24, pixelofs);
  setlelong(buf+28, pixelsize);
  fwrite(buf, 32, 1, fp);
  write_mag_palette_section(fp);
  if (comp) {
    /* write flagA */
    fwrite(flaga, 1, flagAsize, fp);
    /* flagB */
    fwrite(flagb, 1, flagBsize, fp);
    /* write pixel */
    fwrite(pixelbuf, 1, pixelsize, fp);
    free_compress_work();
  } else {
    /* write flagA */
    l = flagAsize;
    bzero(buf, sizeof(buf));
    while (l > 0) {
      i = l < sizeof(buf) ? l : sizeof(buf);
      fwrite(buf, i, 1, fp);
      l -= i;
    }
    /* No need to save flagB */
    /* write pixel */
    for (y = 0; y < imgfileh; y++) {
      p = imgdata+y*imgmaxw;
      for (x = 0; x < imgfilew; x += 4) {
	buf[0] = (*p << 4) | *(p+1);
	buf[1] = (*(p+2) << 4) | *(p+3);
	p += 4;
	fwrite(buf, 2, 1, fp);
      }
    }
  }
  return ferror(fp);
}

int rd_mag(fp)
     FILE *fp;
{
  int w, h;
  
  free_color_all();
  decode_mag(fp, "*noname*");
  get_color(color_buf, mag_palcnt);
  get_requested_colors();
  w = mag_dataw;
  h = mag_datah;
  if (w > imgmaxw) {
    w = imgmaxw;
    message("!\nwidth %d trancate to %d.\n", mag_dataw, w);
  }
  if (h > imgmaxh) {
    h = imgmaxh;
    message("!\nheight %d trancate to %d.\n", mag_datah, h);
  }
  imgupdate(w, h);
  return 0;
}

/* End of file */
