/* $Header: /home/yav/xpx/RCS/message.c,v 1.12 1995/11/23 16:28:39 yav Exp $
 * xpx message
 * written by yav (UHD98984@pcvan.or.jp)
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "headers.h"
#include "xpx.h"
#include "work.h"
#define PUBLIC_MESSAGE_C
#include "extern.h"

#define MSGLINEMAX	16
#define MSGLINESPACE	20
#define MSGWIDTH	400
#define MSGHEIGHT	(MSGLINESPACE*MSGLINEMAX)

char rcsid_message[] = "$Id: message.c,v 1.12 1995/11/23 16:28:39 yav Exp $";

char *errname = NULL;
int errcode = 1;
static Window messagewin;
static MENU msgline[MSGLINEMAX+1];

#ifdef HAVE_STDARG_H
int error(
#if NeedFunctionPrototypes
	  const char *format, ...)
#else
     format)
     char *format;
#endif
{
  va_list ap;
  
  va_start(ap, format);
  if (errname != NULL)
    fprintf(stderr, " %s:", errname);
  vfprintf(stderr, format, ap);
  fprintf(stderr, "\007\n");
  va_end(ap);
  return errcode;
}
#else /* HAVE_STDARG_H */
#ifdef HAVE_VARARGS_H
int error(va_alist)
     va_dcl
{
  va_list ap;
  char *format;
  
  va_start(ap);
  if (errname != NULL)
    fprintf(stderr, " %s: ", errname);
  format = va_arg(ap, char *);
  vfprintf(stderr, format, ap);
  fprintf(stderr, "\007\n");
  va_end(ap);
  return errcode;
}
#else /* HAVE_VARARGS_H */
int error(
#if NeedFunctionPrototypes
	  char *format)
#else
     format)
     char *format;
#endif
{
  if (errname != NULL)
    fprintf(stderr, "\n%s: ", errname);
  vfprintf(stderr, format, &format + 1);
  fprintf(stderr, "\007\n");
  return errcode;
}
#endif /* HAVE_VARARGS_H */
#endif /* HAVE_STDARG_H */

static int current_line = 0;

void map_message_window()
{
  XMapRaised(dsp, messagewin);
}

void message_scroll()
{
  int i;
  char *s;

  s = msgline[0].str;
  for (i = 0; i < MSGLINEMAX-1; i++) {
    msgline[i].str = msgline[i+1].str;
    redraw_window(msgline[i].win);
  }
  *s = '\0';
  msgline[i].str = s;
}

void message_store(str)
     char *str;
{
  char *p;
  
  for (;;) {
    p = strchr(str, '\n');
    if (p != NULL)
      *p++ = '\0';
    strncat(msgline[current_line].str, str, msgline[current_line].strlen);
    redraw_window(msgline[current_line].win);
    if (p == NULL)
      break;
    /* new line */
    if (++current_line >= MSGLINEMAX) {
      message_scroll();
      current_line = MSGLINEMAX-1;
    }
    str = p;
  }
}

#ifdef HAVE_STDARG_H
void message(
#if NeedFunctionPrototypes
	     const char *format, ...)
#else
     format)
     char *format;
#endif
{
  va_list ap;
  char buf[1024];
  
  va_start(ap, format);
  if (*format == '!') {
    map_message_window();
    format++;
  }
  vsprintf(buf, format, ap);
  message_store(buf);
  va_end(ap);
}
#else /* HAVE_STDARG_H */
#ifdef HAVE_VARARGS_H
void message(va_alist)
     va_dcl
{
  va_list ap;
  char *format;
  char buf[1024];
  
  va_start(ap);
  format = va_arg(ap, char *);
  if (*format == '!') {
    map_message_window();
    format++;
  }
  vsprintf(buf, format, ap);
  message_store(buf);
  va_end(ap);
}
#else /* HAVE_VARARGS_H */
int message(
#if NeedFunctionPrototypes
	    char *format)
#else
     format)
     char *format;
#endif
{
  char buf[1024];
  
  if (*format == '!') {
    map_message_window();
    format++;
  }
  vsprintf(buf, format, &format + 1);
  message_store(buf);
}
#endif /* HAVE_VARARGS_H */
#endif /* HAVE_STDARG_H */


int msgev(
#if NeedFunctionPrototypes
	  MENU *,
	  XEvent *
#endif
	  );

int create_message_window()
{
  int i, y;
  MENU *p;
  
  create_toplevel_window(&messagewin, "message", "", MSGWIDTH, MSGHEIGHT);
  XSelectInput(dsp, messagewin, ExposureMask);
  
  p = msgline;
  y = 0;
  for (i = 0; i < MSGLINEMAX; i++) {
    bzero(p, sizeof(*p));
    p->n = i;
    p->y = y;
    p->w = MSGWIDTH+400;
    p->h = MSGLINESPACE;
    p->mode = MenuModeMaskLeft;
    p->im = ExposureMask|ButtonPressMask;
    p->event = msgev;
    p->strlen = 64;
    p++;
    y += MSGLINESPACE;
  }
  p->n = -1;
  menuwin_create(msgline, messagewin, 0, 0);
  return 0;
}

void unmap_message_window()
{
  XUnmapWindow(dsp, messagewin);
}

void close_message_window()
{
  XDestroyWindow(dsp, messagewin);
}

int msgev(p, ev)
     MENU *p;
     XEvent *ev;
{
  switch(ev->type) {
  case ButtonPress:
    unmap_message_window();
    return 1;
  }
  return 0;
}

/* End of file */
