/* $Header: /home/yav/xpx/RCS/sub.c,v 1.30 1996/04/08 18:21:02 yav Exp $
 * xpx menu subroutines
 * written by yav (UHD98984@pcvan.or.jp)
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "headers.h"
#include "xpx.h"
#include "work.h"
#include "icondef.h"
#define PUBLIC_SUB_C
#include "extern.h"

char rcsid_sub[] = "$Id: sub.c,v 1.30 1996/04/08 18:21:02 yav Exp $";

#define MENUTBLMAX 64
static MENU *menutbl[MENUTBLMAX];

void get_drawable_size(dr, wp, hp)
     Drawable dr;
     int *wp;
     int *hp;
{
  Window root;
  int x, y;
  unsigned int width, height, border, depth;
  
  XGetGeometry(dsp, dr, &root, &x, &y, &width, &height, &border, &depth);
  if (wp != NULL)
    *wp = width;
  if (hp != NULL)
    *hp = height;
}

MENU *get_top_menu(p)
     MENU *p;
{
  while (p->n)
    --p;
  return p;
}

#define highlightGC(mode) \
  (((mode) & MenuModeMaskEnter)?(((mode) & MenuModeMaskOn)?gc1:gc0):\
   (((mode) & MenuModeMaskOn)?gc0:gc1))

void menu_highlight(p)
     MENU *p;
{
  XDrawRectangle(dsp, p->win, highlightGC(p->mode), 0, 0, p->w-1, p->h-1);
}

void menuwin_strpos_center(wd, fs, str, xp, yp, lp, ascp, desp, w, h)
     Window wd;
     XFontStruct *fs;
     char *str;
     int *xp;
     int *yp;
     int *lp;
     int *ascp;
     int *desp;
     int w;
     int h;
{
  int len, tw, dir, asc, des;
  XCharStruct overall;
  
  len = strlen(str);
  tw = XTextWidth(fs, str, len);
  XTextExtents(fs, str, 0, &dir, &asc, &des, &overall);
  if (xp != NULL)
    *xp = (w-tw)/2;
  if (yp != NULL)
    *yp = h - (h-(asc+des))/2 - des;
  if (lp != NULL)
    *lp = len;
  if (ascp != NULL)
    *ascp = asc;
  if (desp != NULL)
    *desp = des;
}

#define menu_fill_black(menup) \
  XFillRectangle(dsp,(menup)->win,gc0,0,0,(menup)->w,(menup)->h)

void draw_center_string(wd, w, h, fs, gc, str)
     Window wd;
     int w;
     int h;
     XFontStruct *fs;
     GC gc;
     char *str;
{
  int x, y, len;
  
  menuwin_strpos_center(wd, fs, str, &x, &y, &len, NULL, NULL, w, h);
  XDrawImageString(dsp, wd, gc, x, y, str, len);
}

void menuwin_center(wd, w, h, str)
     Window wd;
     int w;
     int h;
     char *str;
{
  draw_center_string(wd, w, h, fs1, gc0, str);
}

void erase_cursor(p, str)
     MENU *p;
     char *str;
{
  int x, y, len, asc, des;
  
  menuwin_strpos_center(p->win, fs1, str, &x, &y,
			&len, &asc, &des, p->w, p->h);
  if (len)
    x += XTextWidth(fs1, str, p->cursor);
  XClearArea(dsp, p->win, x-2, y-asc-1, 5, asc+3, False);
}

void draw_cursor(
#if NeedFunctionPrototypes
		 MENU *p, char *str)
#else
     p, str)
     MENU *p;
     char *str;
#endif
{
  int x, y, len, asc, des;
  GC gc;
  
  menuwin_strpos_center(p->win, fs1, str, &x, &y,
			&len, &asc, &des, p->w, p->h);
  if (len)
    x += XTextWidth(fs1, str, p->cursor);
  gc = p->mode & MenuModeMaskOn ? gc1 : gc0;
  XDrawLine(dsp, p->win, gc, x, y-asc, x, y);
  XDrawLine(dsp, p->win, gc, x, y, x-2, y+1);
  XDrawLine(dsp, p->win, gc, x, y, x+2, y+1);
  XDrawLine(dsp, p->win, gc, x, y-asc, x-2, y-asc-1);
  XDrawLine(dsp, p->win, gc, x, y-asc, x+2, y-asc-1);
}

void menu_icon_center(p, icon)
     MENU *p;
     int icon;
{
  if (p->mode & MenuModeMaskOn) {
    menu_fill_black(p);
    icon |= ICON_REVERSE;
  }    
  XCopyArea(dsp, icon_pixmap(icon), p->win, gc0,
	    0, 0, icon_width(icon), icon_height(icon),
	    (p->w-icon_width(icon))/2, (p->h-icon_height(icon))/2);
  menu_highlight(p);
}

void menuev_init()
{
  int i;
  
  for (i = 0; i < MENUTBLMAX; i++)
    menutbl[i] = NULL;
}

int set_menutbl(menu)
     MENU *menu;
{
  int i;
  
  for (i = 0; i < MENUTBLMAX; i++) {
    if (menutbl[i] == NULL) {
      menutbl[i] = menu;
      return 0;
    }
  }
  return 1;
}

int unset_menutbl(menu)
     MENU *menu;
{
  int i;
  
  for (i = 0; i < MENUTBLMAX; i++) {
    if (menutbl[i] == menu) {
      menutbl[i] = NULL;
      return 0;
    }
  }
  return 1;
}

void menuwin_create2(
#if NeedFunctionPrototypes
			MENU *menu, Window parent, int x0, int y0, int mapsw)
#else
     menu, parent, x0, y0, mapsw)
     MENU *menu;
     Window parent;
#endif
{
  MENU *p;
  XSetWindowAttributes attr;
  int x, y;
  int parentw, parenth;
  
  get_drawable_size(parent, &parentw, &parenth);
  for (p = menu; p->n >= 0; p++) {
    if (p->strlen && p->str == NULL) {
      p->str = malloc(p->strlen);
      if (p->str == NULL)
	error("menuwin_create: more core!");
      else
	*(p->str) = '\0';
    }
    x = (p->gr==EastGravity||p->gr==NorthEastGravity||p->gr==SouthEastGravity)
      ? parentw - p->w - p->x : p->x;
    y = (p->gr==SouthGravity||p->gr==SouthEastGravity||p->gr==SouthWestGravity)
      ? parenth - p->h - p->y : p->y;
    if (p->gr == CenterGravity) {
      x = (parentw - p->w) / 2 + p->x;
      y = (parenth - p->h) / 2 + p->y;
    }
    p->win = 
      XCreateSimpleWindow(dsp, parent, x0 + x, y0 + y, p->w, p->h,
			  p->mode & MenuModeMaskBorder ? BDW : 0, fg, bg);
    XSelectInput(dsp, p->win, p->im);
    if (p->gr) {
      attr.win_gravity = p->gr;
      XChangeWindowAttributes(dsp, p->win, CWWinGravity, &attr);
    }
    if (mapsw)
      XMapRaised(dsp, p->win);
  }
  set_menutbl(menu);
}

void menuwin_create(
#if NeedFunctionPrototypes
		    MENU *menu, Window parent, int x0, int y0)
#else
     menu, parent, x0, y0)
     MENU *menu;
     Window parent;
#endif
{
  menuwin_create2(menu, parent, x0, y0, 1);
}

void topmenuwin_create(
#if NeedFunctionPrototypes
		       MENU *menu, int x0, int y0)
#else
     menu, parent, x0, y0)
     MENU *menu;
     Window parent;
#endif
{
  menuwin_create2(menu, win, x0, y0, menumode);
}

void menuwin_destroy(menu)
     MENU *menu;
{
  MENU *p;
  
  for (p = menu; p->n >= 0; p++)
    XDestroyWindow(dsp, p->win);
  unset_menutbl(menu);
}

void menuwin_expose_default(p)
     MENU *p;
{
  char *s;
  GC gc;
  
  s = p->strfunc ? p->strfunc(p) : p->str;
  if (s == NULL)
    return;
  if (p->mode & MenuModeMaskOn) {
    menu_fill_black(p);
    gc = gc1;
  } else {
    gc = gc0;
  }
  if (p->mode & MenuModeMaskLeft)
    XDrawImageString(dsp, p->win, gc, 2, p->h-5, s, strlen(s));
  else
    draw_center_string(p->win, p->w, p->h, fs1, gc, s);
  if (p->im & (ButtonPressMask|KeyPressMask))
    menu_highlight(p);
  /* if highlight on , display input cursor */
  if ((p->im & KeyPressMask)&&(p->mode & MenuModeMaskEnter))
    draw_cursor(p, s);
}

void redraw_window(w)
     Window w;
{
  XClearArea(dsp, w, 0, 0, 0, 0, True);
}

void menuev_destroy()
{
  int i;
  
  for (i = MENUTBLMAX-1; i >= 0; --i) {
    if (menutbl[i] != NULL)
      menuwin_destroy(menutbl[i]);
  }
}

void menuev_event_sub(p, ev)
     MENU *p;
     XEvent *ev;
{
  int i;
  short last_mode;
  char *s;
  
  if (p->event != NULL && p->event(p, ev))
    return;
  /* default event operation */
  last_mode = p->mode;
  switch(ev->type) {
  case Expose:
    if (!ev->xexpose.count)
      menuwin_expose_default(p);
    return;
  case GraphicsExpose:
    if (!ev->xgraphicsexpose.count)
      menuwin_expose_default(p);
    return;
  case NoExpose:
    return;
  case ButtonPress:
    if (p->im & ButtonPressMask) {
      if (p->im & ButtonReleaseMask)
	p->mode |= MenuModeMaskOn; /* select on */
      else
	p->mode ^= MenuModeMaskOn; /* select alternate */
    }
    break;
  case ButtonRelease:
    if (p->im & ButtonReleaseMask)
      p->mode &= ~MenuModeMaskOn; /* select off */
    break;
  case EnterNotify:
    if (p->im & (ButtonPressMask|KeyPressMask)) {
      p->mode |= MenuModeMaskEnter; /* highlight on */
      if (!(p->im & KeyPressMask)) {
	menu_highlight(p);
	return;
      }
    }
    break;
  case LeaveNotify:
    if (ev->xcrossing.detail == NotifyInferior)
      return;
    if (p->im & (ButtonPressMask|KeyPressMask)) {
      p->mode &= ~MenuModeMaskEnter; /* highlight off */
      if (!(p->im & KeyPressMask)) {
	menu_highlight(p);
	return;
      }
    }
    break;
  case KeyPress:
    s = p->strfunc ? p->strfunc(p) : p->str;
    erase_cursor(p, s);
    i = edit_string(ev, s, p->strlen, &p->cursor);
    if (i & 2)
      redraw_window(p->win);
    else
      menuwin_expose_default(p);
    return;
  }
  if (p->mode & ~MenuModeMaskEnter == last_mode & ~MenuModeMaskEnter)
    menuwin_expose_default(p);
  else
    redraw_window(p->win);
}

Drawable get_event_window(ev)
     XEvent *ev;
{
  Drawable win;
  
  win = 0;
  switch(ev->type) {
  case Expose:
    win = ev->xexpose.window;
    break;
  case ButtonPress:
    win = ev->xbutton.window;
    break;
  case ButtonRelease:
    win = ev->xbutton.window;
    break;
  case EnterNotify:
    win = ev->xcrossing.window;
    break;
  case LeaveNotify:
    win = ev->xcrossing.window;
    break;
  case KeyPress:
    win = ev->xkey.window;
    break;
  case GraphicsExpose:
    win = ev->xgraphicsexpose.drawable;
    break;
  case NoExpose:
    win = ev->xnoexpose.drawable;
    break;
  case MotionNotify:
    win = ev->xmotion.window;
    break;
  }
  return win;
}

void menuev_event(ev)
     XEvent *ev;
{
  int i;
  MENU *p;
  Drawable win;
  
  win = get_event_window(ev);
  if (!win)
    return;
  for (i = 0; i < MENUTBLMAX; i++) {
    for (p = menutbl[i]; p != NULL && p->n >= 0; p++) {
      if (p->win == win) {
	menuev_event_sub(p, ev);
	return;
      }
    }
  }
}

void gen_default_geometry_string(
#if NeedFunctionPrototypes
				 char *p, int w, int h)
#else
     p, w, h)
     char *p;
     int w;
     int h;
#endif
{
  int rootw, rooth;
  
  get_drawable_size(DefaultRootWindow(dsp), &rootw, &rooth);
  sprintf(p, "%dx%d+%d+%d", w, h, (rootw-w)/2, (rooth-h)/2);
}

int create_toplevel_window(
#if NeedFunctionPrototypes
		       Window *wd,
		       char *name,
		       char *geomstr,
		       int default_width,
		       int default_height)
#else
     wd, name, geomstr, default_width, default_height)
     Window *wd;
     char *name;
     char *geomstr;
     int default_width;
     int default_height;
#endif
{
  int i;
  XSizeHints hint;
  XWMHints wmhint;
  char buf[256];
  
  gen_default_geometry_string(buf, default_width, default_height);
  i = XGeometry(dsp, scr, geomstr, buf, border_width, 1, 1, 0, 0,
		&(hint.x), &(hint.y), &(hint.width), &(hint.height));
#if 0
  if (i & (XValue|YValue))
    hint.flags |= PPosition;
  if (i & (WidthValue|HeightValue))
    hint.flags |= PSize;
#else
  hint.flags = USPosition|USSize;
#endif
  *wd = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), hint.x, hint.y,
			    hint.width, hint.height,
			    border_width, border_color, bg);
  sprintf(buf, "%s %s", myname, name);
  XSetStandardProperties(dsp, *wd, buf, buf, None, oargv, oargc, &hint);
  /* window group hint */
  wmhint.flags = WindowGroupHint;
  wmhint.window_group = win;
  XSetWMHints(dsp, *wd, &wmhint);
  if (toplevelwc < MAXTOPLEVELWINDOW)
    toplevelwin[toplevelwc++] = *wd;
  XSetWindowColormap(dsp, *wd, cmap);
  return 0;
}

static int cursor_type = 0;

void cursor_free(
#if NeedFunctionPrototypes
		 void
#endif
		 )
{
  int i;
  
  if (cursor_type != 0) {
    cursor_type = 0;
    for (i = 0; i < toplevelwc; i++)
      XUndefineCursor(dsp, toplevelwin[i]);
    XDefineCursor(dsp, imgwin, curs_cross);
    XFlush(dsp);
  }
}

void cursor_busy(
#if NeedFunctionPrototypes
		 void
#endif
		 )
{
  int i;
  
  if (cursor_type != 1) {
    cursor_type = 1;
    for (i = 0; i < toplevelwc; i++)
      XDefineCursor(dsp, toplevelwin[i], curs_watch);
    XDefineCursor(dsp, imgwin, curs_watch);
    XFlush(dsp);
  }
}

/* End of file */
